/* MultiViewImage.cpp ======================================================================*

*Important
===================================================
Raytrix Lenslet Converting Version 0.2

Copyright: ONLY Available for ACADEMIC and MPEG Standardization Purposes
Production: Nagoya University


*Description
===================================================
Given a set of Raw video captured by multi-focused plenoptic camera (i.e. Ratrix camera, https://raytrix.de/) 
and its microlens calibration data, the provided source code can render multi-view video.

This tool is implemented by:

Fujii Laboratory
Department of Information and Communication Engineering
Graduate School of Engineering
Nagoya University


*Term of Use
===================================================
Term of Use: ANY kind of academic publication, and report using this tool MUST refer to the following TWO references.

[1] Shu Fujita, Sho Mikawa, Mehrdad Teratani(Panahpourtehrani), Keita Takahashi, and Toshiaki Fujii 
“Extracting Multi-View Images from Multi-Focused Plenoptic Camera,” 
International Workshop on Advanced Image Technology (IWAIT)/International Forum on Medical Imaging in Asia (IFMIA), 
in SPIE Digital Library, Singapore, January 2019

[2] ] Mehrdad Teratani(Panahpourtehrani), Shu Fujita, Wenzhe Ouyang, Keita Takahashi, Toshiaki Fujii, 
"A 3-D Imaging System Using Multi-Focused Plenoptic Camera and Tensor Display," 
2018 International Conference on 3D Immersion (IC3D2018) in IEEE Xplore Digital Library, Brussels, Belgium, December 2018 

===================================================
* Class for generating raytrix multi-view image
*
* @file		RenderingRxMV > MultiViewImage.h
* @author	S. Mikawa, S. Fujita, and W. Ouyang
* @date		2019/04/17
*======================================================================*/

#include "MultiViewImage.h"

int compare_uchar(const void *a, const void *b)
{
	return *(uchar*)a - *(uchar*)b;
}

//Constructor
MultiViewImage::MultiViewImage(MicroLens ml, const int imageNum, int skipped_number_of_pixel_around_ML)
{
	//Set Micro Lens Parameter
	this->ml = ml;
	ml.showParam();

	//Generate Base Mask Image
	maskImage = Mat::zeros(Size((int)ml.diameter, (int)ml.diameter), CV_8UC1);
	generateMask(maskImage);

	//Make Mask Image Border
	Mat maskBorderImage;
	copyMakeBorder(maskImage, maskBorderImage, 1, 1, 1, 1, BORDER_CONSTANT, Scalar::all(0));
	cv::Laplacian(maskBorderImage, maskLapImage, maskBorderImage.depth(), 3, 1, 127, BORDER_DEFAULT);
	compare(maskLapImage, 127, maskLapImage, CMP_NE);

	//Generate Patch Size List and Allocation patchSize Memory
	subImageOverlap = skipped_number_of_pixel_around_ML;
	/*subImageOverlap: it sets the pixel distance between the maximum patchsize and the border of microlens.  
    Generally, if you set this value bigger, the generating multi-view images will be more accurate but have 
    a lower spatial resolution; if you set this value smaller,  the generating multi-view images will have 
    a higher spatial resolution but more serious artifacts. We recommend you to set this value to 1.
    */

    //Patch Size Interval
	patchInterval = 1;	//1 or 2
	for (int patch = 3; patch <= (int)(ml.diameter - subImageOverlap * 2); patch += patchInterval)
		patchSizeList.push_back(patch);
	for (int type = TYPE1; type < TYPE_ALL; type++)
	{
		patchSize[type] = Mat::ones(Size(ml.num.x, ml.num.y), CV_8UC1);
		edgeStd[type] = Mat::ones(Size(ml.num.x, ml.num.y), CV_32FC1);
		for (int i = 0; i < DIVISION_NUM; i++)
		{
			dividedPatchSize[type][i] = Mat::ones(Size(ml.num.x, ml.num.y), CV_8UC1);
		}
	}

	//Set Stereo Matching Parameter
	int minDisparity = (int)(patchSizeList.front() * 0.5);
	int maxDisparity = (int)(patchSizeList.back() * 0.5);
	Size blockSize(5, 5);
	rsm = new RaytrixStereoMatching(ml, minDisparity, maxDisparity, blockSize);

	//Allocation multiViewImage Memory
	centerView = Point((int)(imageNum / 2), (int)(imageNum / 2));
	Point2d lensCenter;
	lensType rightLensType = ml.getRefLensType(0, RIGHT);
	lensCenter.x = ml.subGridRefPos[0][rightLensType].x + ml.diameter * 3 * (ml.num.x - 1);
	lensCenter.y = ml.subGridRefPos[0][rightLensType].y + ml.radius * sqrt(3) * (ml.num.y - 1);
	multiViewImageSize.width = (int)floor(ml.subGridRefPos[0][rightLensType].x + (lensCenter.x - ml.subGridRefPos[0][rightLensType].x) * ((ml.radius - subImageOverlap) / ml.diameter) + ml.diameter + centerView.x);
	multiViewImageSize.height = (int)floor(ml.subGridRefPos[0][rightLensType].y + (lensCenter.y - ml.subGridRefPos[0][rightLensType].y) * ((ml.radius - subImageOverlap) / ml.diameter) + ml.diameter + centerView.y);

	this->imageNum = imageNum;
	Point view;
	for (int type = TYPE1; type < TYPE_ALL; type++)
	{
		multiViewImage[type].resize(imageNum);
		for (view.x = 0; view.x < imageNum; view.x++)
		{
			multiViewImage[type][view.x].resize(imageNum);
		}
	}
}

//Destructor
MultiViewImage::~MultiViewImage()
{
}

int MultiViewImage::getPatchListIndex(int patchSize)
{
	return (int)((patchSize - patchSizeList.front()) / patchInterval);
}

void MultiViewImage::generateMask(Mat& maskImage)
{
	int shift = 4;
	Point2d center = (1 << shift) * Point2d(maskImage.cols * 0.5, maskImage.rows * 0.5);
	circle(maskImage, center, (int)center.x, Scalar::all(255), -1, 8, shift);

	//Generate Patch which is Divided into 6	
	for (int i = 0; i < DIVISION_NUM; i++)
	{
		dividedMaskImage[i] = Mat::zeros(maskImage.size(), maskImage.type());
		int angle = 360 / DIVISION_NUM * i;
		Size2d axes = Size2d((1 << shift) * maskImage.cols * 0.5, (1 << shift) * maskImage.rows * 0.5);
		ellipse(dividedMaskImage[i], center, axes, 0, angle, angle + (360 / DIVISION_NUM), Scalar::all(255), -1, 8, shift);
	}
}

void MultiViewImage::generateResizeRawImage()
{
	//Make Border to Prevent Access to the Outside Raw Image Area
	copyMakeBorder(rawImage, rawImage, centerView.y, centerView.y, centerView.x, centerView.x, BORDER_CONSTANT, Scalar::all(0));
	ml.center += (Point2d)centerView;
	for (int updown = 0; updown < 2; updown++)
		for (int type = 0; type < 3; type++)
			ml.subGridRefPos[updown][type] += (Point2d)centerView;

	//resizeRawImageList[0] is Gray Image, resizeRawImageList[0] is Color Image
	if (rawImage.channels() == 1)
		resizeRawImageList.resize(1);
	else if (rawImage.channels() == 3)
		resizeRawImageList.resize(2);

	int patchCnt = 0;
	float patchRatio;
	for (int i = 0; i < resizeRawImageList.size(); i++)
		resizeRawImageList[i].resize(patchSizeList.size() + subImageOverlap * 2);
	for (int i = 1; patchCnt < resizeRawImageList[0].size(); patchCnt++)
	{
		if (patchCnt < patchSizeList.size())
			patchRatio = (float)maskImage.cols / patchSizeList[patchCnt];
		else
		{
			patchRatio = (float)maskImage.cols / (patchSizeList.back() + patchInterval * i);
			i++;
		}
		if (rawImage.channels() == 1)
		{
			//cout << rawImage.size() << endl;
			//cout << Size(int(rawImage.cols * patchRatio), int(rawImage.rows * patchRatio)) << endl;
			//cout << Size(int(rawImage.cols * patchRatio), int(rawImage.rows * patchRatio)).area() << endl << endl;
			resize(rawImage, resizeRawImageList[0][patchCnt], Size(int(rawImage.cols * patchRatio), int(rawImage.rows * patchRatio)), 0.0, 0.0, INTER_LINEAR);
		}
		else if (rawImage.channels() == 3)
		{
			resize(rawImage, resizeRawImageList[1][patchCnt], Size(int(rawImage.cols * patchRatio), int(rawImage.rows * patchRatio)), 0.0, 0.0, INTER_LINEAR);
			cvtColor(resizeRawImageList[1][patchCnt], resizeRawImageList[0][patchCnt], CV_RGB2GRAY);
		}
	}
}

bool MultiViewImage::checkRoiRange(Mat& srcImage, Rect roi)
{
	if ((srcImage.cols < (roi.x + roi.width)) ||
		(srcImage.rows < (roi.y + roi.height)) ||
		(0 > roi.x) ||
		(0 > roi.y))
	{
		cout << "Out of Roi Range: (" << srcImage.cols << "," << srcImage.rows << ") -> " << roi << endl;
		return false;
	}
	return true;
}

void MultiViewImage::alphaBlend(Mat& srcImage, Mat& destImage, Mat& maskImage, double alpha)
{
	if (srcImage.depth() == CV_8U)
	{
		if (srcImage.channels() == 1)
		{
			for (int y = 0; y < srcImage.rows; y++)
			{
				uchar* mptr = maskImage.ptr<uchar>(y);
				uchar* sptr = srcImage.ptr<uchar>(y);
				uchar* dptr = destImage.ptr<uchar>(y);

				for (int x = 0; x < srcImage.cols; x++)
				{
					if (mptr[x] == 255)
					{
						if (dptr[x] == 0)
						{
							dptr[x] = sptr[x];
						}
						else
						{
							dptr[x] = (int)(dptr[x] * alpha + sptr[x] * (1 - alpha));
						}
					}
				}
			}
		}
		else if (srcImage.channels() == 3)
		{
			for (int ch = 0; ch < srcImage.channels(); ch++)
			{
				for (int y = 0; y < srcImage.rows; y++)
				{
					uchar* mptr = maskImage.ptr<uchar>(y);
					Vec3b* sptr = srcImage.ptr<Vec3b>(y);
					Vec3b* dptr = destImage.ptr<Vec3b>(y);

					for (int x = 0; x < srcImage.cols; x++)
					{
						if (mptr[x] == 255)
						{
							if (dptr[x][ch] == 0)
							{
								dptr[x][ch] = sptr[x][ch];
							}
							else
							{
								dptr[x][ch] = (int)(dptr[x][ch] * alpha + sptr[x][ch] * (1 - alpha));
							}
						}
					}
				}
			}
		}
	}
	else if (srcImage.depth() == CV_32F)
	{
		if (srcImage.channels() == 1)
		{
			for (int y = 0; y < srcImage.rows; y++)
			{
				uchar* mptr = maskImage.ptr<uchar>(y);
				float* sptr = srcImage.ptr<float>(y);
				float* dptr = destImage.ptr<float>(y);

				for (int x = 0; x < srcImage.cols; x++)
				{
					if (mptr[x] == 255)
					{
						if (dptr[x] == 0)
						{
							dptr[x] = sptr[x];
						}
						else
						{
							dptr[x] = dptr[x] * alpha + sptr[x] * (1 - alpha);
						}
					}
				}
			}
		}
		else if (srcImage.channels() == 3)
		{
			for (int ch = 0; ch < srcImage.channels(); ch++)
			{
				for (int y = 0; y < srcImage.rows; y++)
				{
					uchar* mptr = maskImage.ptr<uchar>(y);
					Vec3f* sptr = srcImage.ptr<Vec3f>(y);
					Vec3f* dptr = destImage.ptr<Vec3f>(y);

					for (int x = 0; x < srcImage.cols; x++)
					{
						if (mptr[x] == 255)
						{
							if (dptr[x][ch] == 0)
							{
								dptr[x][ch] = sptr[x][ch];
							}
							else
							{
								dptr[x][ch] = dptr[x][ch] * alpha + sptr[x][ch] * (1 - alpha);
							}
						}
					}
				}
			}
		}
	}
}

void MultiViewImage::getRoiImage(Mat& srcImage, Mat& destImage, Mat& maskImage, Rect roi)
{
	if (checkRoiRange(srcImage, roi))
	{
		Mat roiImage = srcImage(roi);
		roiImage.copyTo(destImage, maskImage);
	}
}

void MultiViewImage::setRoiImage(Mat& srcImage, Mat& destImage, Mat& maskImage, Rect roi)
{
	if (checkRoiRange(destImage, roi))
	{
		Mat roiImage = destImage(roi);
		alphaBlend(srcImage, roiImage, maskImage, 0.5);
		//srcImage.copyTo(roiImage, maskImage);
	}
}

void MultiViewImage::patchVisualization(lensType type)
{
	//patchMap	->		Front:	Small Patch Size
	//					Back:	Big Patch Size
	//patchMapImage ->	Front:	Light
	//					Back:	Dark
	for (int t = TYPE1; t < TYPE_ALL; t++)
	{
		patchVisualizationMap[t] = 255 - 255 * patchMap[t] / patchSizeList.back();
		if (t == ml.getRefLensType(0, LEFT))
			getShiftImage(patchVisualizationMap[t], patchVisualizationMap[t], Point(lensTypeOffset, 0));
		else if (t == ml.getRefLensType(0, RIGHT))
			getShiftImage(patchVisualizationMap[t], patchVisualizationMap[t], Point(-lensTypeOffset, 0));
	}
	for (int t = TYPE1; t < TYPE_ALL; t++)
	{
		double minv, maxv;
		minMaxLoc(edgeStdMap[type], &minv, &maxv);
		edgeStdMap[t].convertTo(edgeStdVisualizationMap[t], CV_8U, 255 / maxv);
	}
	if (type != TYPE_ALL)
		cv::imshow("patchMapVisualization", patchVisualizationMap[type]);
}

void MultiViewImage::generatePatchMap(int mode)
{
	Point lens;
	/*
	if (mode == DEPTH_BASE_V3)
	{
		for (int type = TYPE1; type < TYPE_ALL; type++)
		{
			patchMap[type] = Mat::zeros(multiViewImageSize, CV_8UC1);
			for (lens.y = 0; lens.y < ml.num.y; lens.y++)
			{
				for (lens.x = 0; lens.x < ml.num.x; lens.x++)
				{
					Point2d lensCenter;
					lensCenter.x = ml.subGridRefPos[lens.y % 2][type].x + ml.diameter * 3 * lens.x;
					lensCenter.y = ml.subGridRefPos[0][type].y + ml.radius * sqrt(3) * lens.y;
					lensCenter.x = ml.subGridRefPos[0][type].x + (lensCenter.x - ml.subGridRefPos[0][type].x) * ((ml.radius - subImageOverlap) / ml.diameter);
					lensCenter.y = ml.subGridRefPos[0][type].y + (lensCenter.y - ml.subGridRefPos[0][type].y) * ((ml.radius - subImageOverlap) / ml.diameter);

					for (int i = 0; i < DIVISION_NUM; i++)
					{
						Mat subImage = Mat::zeros(maskImage.size(), CV_8UC1);
						int shift = 4;
						int angle = 360 / DIVISION_NUM * i;
						Point2d center = (1 << shift) * Point2d(maskImage.cols * 0.5, maskImage.rows * 0.5);
						Size2d axes = Size2d((1 << shift) * maskImage.cols * 0.5, (1 << shift) * maskImage.rows * 0.5);
						ellipse(subImage, center, axes, 0, angle, angle + (360 / DIVISION_NUM), Scalar::all(dividedPatchSize[type][i].at<uchar>(lens.y, lens.x)), -1, 8, shift);

						Rect2d setRoi(lensCenter.x - maskImage.cols * 0.5, lensCenter.y - maskImage.rows * 0.5, maskImage.cols, maskImage.rows);
						setRoiImage(subImage, patchMap[type], dividedMaskImage[i], setRoi);
					}
				}
			}
		}
	}
	*/
	//else
	{
		for (int type = TYPE1; type < TYPE_ALL; type++)
		{
			patchMap[type] = Mat::zeros(multiViewImageSize, CV_8UC1);
			edgeStdMap[type] = Mat::zeros(multiViewImageSize, CV_32FC1);
			for (lens.y = 0; lens.y < ml.num.y; lens.y++)
			{
				for (lens.x = 0; lens.x < ml.num.x; lens.x++)
				{
					//Convert Patch Size into Circle Image
					Mat subImage(maskImage.size(), CV_8UC1, Scalar::all(patchSize[type].at<uchar>(lens.y, lens.x)));
					Mat subStd(maskImage.size(), CV_32FC1, Scalar::all(edgeStd[type].at<float>(lens.y, lens.x)));

					Point2d lensCenter;
					lensCenter.x = ml.subGridRefPos[lens.y % 2][type].x + ml.diameter * 3 * lens.x;
					lensCenter.y = ml.subGridRefPos[0][type].y + ml.radius * sqrt(3) * lens.y;
					lensCenter.x = ml.subGridRefPos[0][type].x + (lensCenter.x - ml.subGridRefPos[0][type].x) * ((ml.radius - subImageOverlap) / ml.diameter);
					lensCenter.y = ml.subGridRefPos[0][type].y + (lensCenter.y - ml.subGridRefPos[0][type].y) * ((ml.radius - subImageOverlap) / ml.diameter);

					Rect2d setRoi(lensCenter.x - maskImage.cols * 0.5, lensCenter.y - maskImage.rows * 0.5, maskImage.cols, maskImage.rows);
					setRoiImage(subImage, patchMap[type], maskImage, setRoi);
					setRoiImage(subStd, edgeStdMap[type], maskImage, setRoi);
				}
			}
		}
	}
}

void MultiViewImage::getShiftImage(Mat& srcImage, Mat& destImage, Point shift)
{
	//shift.x is Positive Number -> Shift to Right
	//shift.y is Positive Number -> Shift to Down

	Mat tempImage(srcImage.size(), srcImage.type());

	const int ch = srcImage.channels();
	const int step = srcImage.cols * ch;
	const int height = srcImage.rows;
	shift.x *= ch;
	shift.y *= ch;

	if (srcImage.depth() == CV_8U)
	{
		if (shift.x >= 0)	//RIGHT Shift
		{
			uchar* s = srcImage.ptr<uchar>(0);
			uchar* d = tempImage.ptr<uchar>(0);

			for (int j = 0; j < height; j++, s += step, d += step)
			{
				const uchar v = s[0];
				memset(d, v, shift.x);
				memcpy(d + shift.x, s, step - shift.x);
			}
		}
		else				//LEFT Shift
		{
			uchar* s = srcImage.ptr<uchar>(0);
			uchar* d = tempImage.ptr<uchar>(0);

			for (int j = 0; j < height; j++, s += step, d += step)
			{
				const uchar v = s[step - 1];
				memcpy(d, s - shift.x, step + shift.x);
				memset(d + step + shift.x, v, -shift.x);
			}
		}

		if (shift.y >= 0)	//Down Shift
		{
			uchar* s = tempImage.ptr<uchar>(0);
			uchar* d = destImage.ptr<uchar>(0);

			for (int j = 0; j < shift.y; j++, d += step)
				memcpy(d, s, step);
			memcpy(d, s, step * (height - shift.y));
		}
		else				//Up Shift
		{
			uchar* s = tempImage.ptr<uchar>(0);
			uchar* d = destImage.ptr<uchar>(0);

			memcpy(d, s - step * shift.y, step * (height + shift.y));
			s += step * (height - 1);
			d += step * (height + shift.y);
			for (int j = shift.y; j < 0; j++, d += step)
				memcpy(d, s, step);
		}
	}
	else if (srcImage.depth() == CV_32F)
	{
		if (shift.x >= 0)	//RIGHT Shift
		{
			float* s = srcImage.ptr<float>(0);
			float* d = tempImage.ptr<float>(0);

			for (int j = 0; j < height; j++, s += step, d += step)
			{
				const float v = s[0];
				memset(d, v, shift.x * sizeof(float));
				memcpy(d + shift.x, s, (step - shift.x) * sizeof(float));
			}
		}
		else				//LEFT Shift
		{
			float* s = srcImage.ptr<float>(0);
			float* d = tempImage.ptr<float>(0);

			for (int j = 0; j < height; j++, s += step, d += step)
			{
				const float v = s[step - 1];
				memcpy(d, s - shift.x, (step + shift.x) * sizeof(float));
				memset(d + step + shift.x, v, -shift.x * sizeof(float));
			}
		}

		if (shift.y >= 0)	//Down Shift
		{
			float* s = tempImage.ptr<float>(0);
			float* d = destImage.ptr<float>(0);

			for (int j = 0; j < shift.y; j++, d += step)
				memcpy(d, s, step * sizeof(float));
			memcpy(d, s, step * (height - shift.y) * sizeof(float));
		}
		else				//Up Shift
		{
			float* s = tempImage.ptr<float>(0);
			float* d = destImage.ptr<float>(0);

			memcpy(d, s - step * shift.y, step * (height + shift.y) * sizeof(float));
			s += step * (height - 1);
			d += step * (height + shift.y);
			for (int j = shift.y; j < 0; j++, d += step)
				memcpy(d, s, step * sizeof(float));
		}
	}
}

void MultiViewImage::laplacianBaseEstimation(bool isSpecificView, Point view)
{
	Point firstView = view, endView(imageNum, imageNum), lens;
	if (isSpecificView)
		endView = view + Point(1, 1);

	vector<Mat> edgeSumMaps(patchSizeList.size());
	Mat multiViewLapImage, multiViewLapRoiImage, resultImage, resultRoiImage;

	for (int type = TYPE1; type < TYPE_ALL; type++)
	{
		//Calculate Edge Cost
		for (int patchCnt = 0; patchCnt < patchSizeList.size(); patchCnt++)
		{
			Mat edgeSumMap = Mat::zeros(Size(ml.num.x, ml.num.y), CV_64FC1);

			for (view.y = firstView.y; view.y < endView.y; view.y++)
			{
				for (view.x = firstView.x; view.x < endView.x; view.x++)
				{
					rendering(true, true, patchSizeList[patchCnt], view);
					//Laplacian(multiViewImage[type][view.x][view.y], multiViewLapImage, multiViewImage[type][view.x][view.y].depth(), 1, 3, 0, BORDER_DEFAULT);
					cv::Laplacian(multiViewImage[type][view.x][view.y], multiViewLapImage, CV_32F, 1, 1, 0, BORDER_DEFAULT);
					resultImage = Mat::zeros(multiViewImage[type][view.x][view.y].size(), CV_8UC1);

					absdiff(multiViewLapImage, 0, multiViewLapImage);

					for (lens.y = 0; lens.y < ml.num.y; lens.y++)
					{
						for (lens.x = 0; lens.x < ml.num.x; lens.x++)
						{
							Point2d lensCenter;
							lensCenter.x = ml.subGridRefPos[lens.y % 2][type].x + ml.diameter * 3 * lens.x + (view.x - centerView.x);
							lensCenter.y = ml.subGridRefPos[0][type].y + ml.radius * sqrt(3) * lens.y + (view.y - centerView.y);
							lensCenter.x = ml.subGridRefPos[0][type].x + (lensCenter.x - ml.subGridRefPos[0][type].x) * ((ml.radius - subImageOverlap) / ml.diameter);
							lensCenter.y = ml.subGridRefPos[0][type].y + (lensCenter.y - ml.subGridRefPos[0][type].y) * ((ml.radius - subImageOverlap) / ml.diameter);
							Rect2d roi(lensCenter.x - maskLapImage.cols * 0.5, lensCenter.y - maskLapImage.rows * 0.5, maskLapImage.cols, maskLapImage.rows);

							if (checkRoiRange(multiViewLapImage, roi))
							{
								multiViewLapRoiImage = multiViewLapImage(roi);
								resultRoiImage = resultImage(roi);
								multiViewLapRoiImage.copyTo(resultRoiImage, maskLapImage);
								edgeSumMap.at<double>(lens.y, lens.x) += (double)sum(resultRoiImage)[0] / sum(maskLapImage)[0] * 100.0;
							}
						}
					}
				}
			}
			//GaussianBlur(edgeSumMap, edgeSumMap, Size(5, 5), 2.f / 3.f);
			edgeSumMaps[patchCnt] = edgeSumMap.clone();
		}

		//Estimate Optimal Patch
		for (lens.y = 0; lens.y < ml.num.y; lens.y++)
		{
			for (lens.x = 0; lens.x < ml.num.x; lens.x++)
			{
				double min = 100 * (endView.x - firstView.x) * (endView.y - firstView.y);
				double mean = 0;
				for (int patchCnt = 0; patchCnt < patchSizeList.size(); patchCnt++)
				{
					double value = (double)edgeSumMaps[patchCnt].at<double>(lens.y, lens.x);
					mean += value / patchSizeList.size();
					if (min > value)
					{
						min = value;
						patchSize[type].at<uchar>(lens.y, lens.x) = patchSizeList[patchCnt];
					}

					if (true)
					{
						Mat vals(Size((int)patchSizeList.size(), 1), CV_64F);
						for (int patchCnt = 0; patchCnt < patchSizeList.size(); patchCnt++)
							vals.at<double>(patchCnt) = edgeSumMaps[patchCnt].at<double>(lens.y, lens.x);
						Scalar mean, stddev;
						cv::meanStdDev(vals, mean, stddev);
						int n = (int)patchSizeList.size();
						float scale = 1;
						if (stddev[0] > 255)
							edgeStd[type].at<float>(lens.y, lens.x) = 255;
						else
						edgeStd[type].at<float>(lens.y, lens.x) = stddev[0] * scale;
					}
				}
			}
		}
		//medianBlur(patchSize[type], patchSize[type], 3);
	}
}

void MultiViewImage::depthBaseEstimationV1()
{
	//Generate Disparity Image
	Mat disparityImage;
	rsm->operator()(rawImage, disparityImage);

	Point lens;
	Point2d lensCenter;

	//Optimal Patch Size = Largest Number of Patch Size in the Disparity Sub Image
	//Calculating cross correlation of center portion of a microlens image with similarly-sized 
	//patches in the neighbouring microlens.
	//Please refer follwing paper for the details,
	//T. Georgiev and A. Lumsdaine, "Reducing Plenoptic Camera Artifacts", Computer Graphics forum, 2010.
	for (int type = TYPE1; type < TYPE_ALL; type++)
	{
		for (lens.y = 0; lens.y < ml.num.y; lens.y++)
		{
			for (lens.x = 0; lens.x < ml.num.x; lens.x++)
			{
				vector<int> disparityTable((int)ml.diameter, 0);

				lensCenter.x = ml.subGridRefPos[lens.y % 2][type].x + ml.diameter * 3 * lens.x;
				lensCenter.y = ml.subGridRefPos[0][type].y + ml.radius * sqrt(3) * lens.y;

				Mat subImage(maskImage.size(), rawImage.type(), Scalar::all(0));
				Rect2d getRoi(lensCenter.x - maskImage.cols * 0.5, lensCenter.y - maskImage.rows * 0.5, maskImage.cols, maskImage.rows);
				getRoiImage(disparityImage, subImage, maskImage, getRoi);

				for (int y = 0; y < subImage.rows; y++)
					for (int x = 0; x < subImage.cols; x++)
						disparityTable[subImage.at<uchar>(y, x)]++;

				//where Disparity:0 = Mask
				vector<int>::iterator maxIter = max_element(disparityTable.begin() + 1, disparityTable.end());
				int maxIndex = (int)distance(disparityTable.begin(), maxIter);

				if (maxIndex % 2)
					patchSize[type].at<uchar>(lens.y, lens.x) = maxIndex * 2 + 1;
				else
					patchSize[type].at<uchar>(lens.y, lens.x) = maxIndex * 2;
			}
		}
	}
}

void MultiViewImage::depthBaseEstimationV2()
{
	//Generate Disparity Image
	Mat disparityImage;
	rsm->operator()(rawImage, disparityImage);

	//Mat disparityVisualizationImage = disparityImage.clone();
	//disparityVisualizationImage *= 16;
	//imshow("disparityImage", disparityVisualizationImage);

	Point lens;
	Point2d lensCenter;
	Mat rawLapImage;
	Mat dispSubImage(maskImage.size(), rawImage.type(), Scalar::all(0));
	Mat rawLabSubImage = dispSubImage.clone();
	cv::Laplacian(rawImage, rawLapImage, rawImage.depth(), 1, 3, 0, BORDER_DEFAULT);

	//Optimal Patch Size = Largest Number of Patch Size in the Disparity Sub Image (a part of the Edge)
	//Similar with depthBaseEstimationV1, but a Laplacian operation is applied to rawImage.
	for (int type = TYPE1; type < TYPE_ALL; type++)
	{
		for (lens.y = 0; lens.y < ml.num.y; lens.y++)
		{
			for (lens.x = 0; lens.x < ml.num.x; lens.x++)
			{
				vector<int> disparityTable((int)ml.diameter, 0);

				lensCenter.x = ml.subGridRefPos[lens.y % 2][type].x + ml.diameter * 3 * lens.x;
				lensCenter.y = ml.subGridRefPos[0][type].y + ml.radius * sqrt(3) * lens.y;

				Rect2d getRoi(lensCenter.x - maskImage.cols * 0.5, lensCenter.y - maskImage.rows * 0.5, maskImage.cols, maskImage.rows);
				getRoiImage(disparityImage, dispSubImage, maskImage, getRoi);
				getRoiImage(rawLapImage, rawLabSubImage, maskImage, getRoi);

				for (int y = 0; y < dispSubImage.rows; y++)
					for (int x = 0; x < dispSubImage.cols; x++)
						disparityTable[dispSubImage.at<uchar>(y, x)] += rawLabSubImage.at<uchar>(y, x);

				//where Disparity:0 = Mask
				vector<int>::iterator maxIter = max_element(disparityTable.begin() + 1, disparityTable.end());
				int maxIndex = (int)distance(disparityTable.begin(), maxIter);

				if (maxIndex % 2)
					patchSize[type].at<uchar>(lens.y, lens.x) = maxIndex * 2 + 1;
				else
					patchSize[type].at<uchar>(lens.y, lens.x) = maxIndex * 2;
			}
		}
	}
}

/*
void MultiViewImage::depthBaseEstimationV3()
{
	//Generate Disparity Image
	Mat disparityImage;
	rsm->operator()(rawImage, disparityImage);

	Point view, lens;
	Point2d lensCenter, resultLensCenter;
	int patch;

	//Patch is Divided into 6 Portion and Calculate Optimal Patch and Rendering
	for (int type = TYPE1; type < TYPE_ALL; type++)
	{
		for (view.x = 0; view.x < imageNum; view.x++)
		{
			for (view.y = 0; view.y < imageNum; view.y++)
			{
				multiViewImage[type][view.x][view.y] = Mat::zeros(multiViewImageSize, CV_8UC1);

				for (lens.y = 0; lens.y < ml.num.y; lens.y++)
				{
					for (lens.x = 0; lens.x < ml.num.x; lens.x++)
					{
						vector<int> disparityTable[DIVISION_NUM];

						//Set Raw Image Lens Center Coordinate
						lensCenter.x = ml.subGridRefPos[lens.y % 2][type].x + ml.diameter * 3 * lens.x + (view.x - centerView.x);
						lensCenter.y = ml.subGridRefPos[0][type].y + ml.radius * sqrt(3) * lens.y + (view.y - centerView.y);

						//Extract Sub-Image from Disparity Image
						Mat disparitySubImage(maskImage.size(), rawImage.type(), Scalar::all(0));
						Rect2d getRoi(lensCenter.x - maskImage.cols * 0.5, lensCenter.y - maskImage.rows * 0.5, maskImage.cols, maskImage.rows);
						getRoiImage(disparityImage, disparitySubImage, maskImage, getRoi);

						//Computation Each Portion
						for (int i = 0; i < DIVISION_NUM; i++)
						{
							//Disparity Aggregation
							disparityTable[i].resize((int)ml.diameter, 0);
							for (int y = 0; y < disparitySubImage.rows; y++)
								for (int x = 0; x < disparitySubImage.cols; x++)
									if (dividedMaskImage[i].at<uchar>(y, x))
										disparityTable[i][disparitySubImage.at<uchar>(y, x)]++;

							//Get Patch Size of One Portion
							vector<int>::iterator maxIter = max_element(disparityTable[i].begin() + 1, disparityTable[i].end());
							int maxIndex = (int)distance(disparityTable[i].begin(), maxIter);
							if (maxIndex % 2)
								patch = maxIndex * 2 + 1;
							else
								patch = maxIndex * 2;
							if ((view.x == centerView.x) && (view.y == centerView.y))
								dividedPatchSize[type][i].at<uchar>(lens.y, lens.x) = patch;

							//Get Sub Image from Raw Image of Optimal Size
							patch += subImageOverlap * 2;
							float patchRatio = (float)maskImage.cols / patch;
							Mat subImage(maskImage.size(), CV_8UC1, Scalar::all(255));
							Rect2d getRoi(lensCenter.x * patchRatio - maskImage.cols * 0.5, lensCenter.y * patchRatio - maskImage.rows * 0.5, maskImage.cols, maskImage.rows);
							getRoiImage(resizeRawImageList[0][getPatchListIndex(patch)], subImage, dividedMaskImage[i], getRoi);

							//Set Sub Image
							resultLensCenter.x = ml.subGridRefPos[0][type].x + (lensCenter.x - ml.subGridRefPos[0][type].x) * ((ml.radius - subImageOverlap) / ml.diameter);
							resultLensCenter.y = ml.subGridRefPos[0][type].y + (lensCenter.y - ml.subGridRefPos[0][type].y) * ((ml.radius - subImageOverlap) / ml.diameter);
							Rect2d setRoi(resultLensCenter.x - maskImage.cols * 0.5, resultLensCenter.y - maskImage.rows * 0.5, maskImage.cols, maskImage.rows);
							setRoiImage(subImage, multiViewImage[type][view.x][view.y], dividedMaskImage[i], setRoi);
						}
					}
				}
			}
		}
	}
}
*/

void MultiViewImage::rendering(bool isSpecificPatchSize, bool isSpecificView, int patch, Point view)
{
	Point firstView = view, endView(imageNum, imageNum);
	if (isSpecificView)
		endView = view + Point(1, 1);

	float patchRatio;
	Point lens;
	Point2d lensCenter;

	if (isSpecificPatchSize)
		patch += subImageOverlap * 2;

	for (int type = TYPE1; type < TYPE_ALL; type++)
	{
		for (view.y = firstView.y; view.y < endView.y; view.y++)
		{
			for (view.x = firstView.x; view.x < endView.x; view.x++)
			{
				if (isSpecificView || (rawImage.channels() == 1))
					multiViewImage[type][view.x][view.y] = Mat::zeros(multiViewImageSize, CV_8UC1);
				else
					multiViewImage[type][view.x][view.y] = Mat::zeros(multiViewImageSize, CV_8UC3);

				for (lens.y = 0; lens.y < ml.num.y; lens.y++)
				{
					for (lens.x = 0; lens.x < ml.num.x; lens.x++)
					{
						if (!isSpecificPatchSize)
							patch = patchSize[type].at<uchar>(lens.y, lens.x) + subImageOverlap * 2;
						patchRatio = (float)maskImage.cols / patch;
						lensCenter.x = ml.subGridRefPos[lens.y % 2][type].x + ml.diameter * 3 * lens.x + (view.x - centerView.x);
						lensCenter.y = ml.subGridRefPos[0][type].y + ml.radius * sqrt(3) * lens.y + (view.y - centerView.y);

						//Get Sub Image from Raw Image of Optimal Size
						Mat subImage(maskImage.size(), rawImage.type(), Scalar::all(255));
						Rect2d getRoi(lensCenter.x * patchRatio - maskImage.cols * 0.5, lensCenter.y * patchRatio - maskImage.rows * 0.5, maskImage.cols, maskImage.rows);
						if (isSpecificView || (rawImage.channels() == 1))
							getRoiImage(resizeRawImageList[0][getPatchListIndex(patch)], subImage, maskImage, getRoi);
						else
							getRoiImage(resizeRawImageList[1][getPatchListIndex(patch)], subImage, maskImage, getRoi);

						//Set Sub Image
						lensCenter.x = ml.subGridRefPos[0][type].x + (lensCenter.x - ml.subGridRefPos[0][type].x) * ((ml.radius - subImageOverlap) / ml.diameter);
						lensCenter.y = ml.subGridRefPos[0][type].y + (lensCenter.y - ml.subGridRefPos[0][type].y) * ((ml.radius - subImageOverlap) / ml.diameter);
						Rect2d setRoi(lensCenter.x - maskImage.cols * 0.5, lensCenter.y - maskImage.rows * 0.5, maskImage.cols, maskImage.rows);
						setRoiImage(subImage, multiViewImage[type][view.x][view.y], maskImage, setRoi);
					}
				}
			}
		}
	}
}

void MultiViewImage::shiftMultiViewImage()
{
	lensType leftLensType = ml.getRefLensType(0, LEFT);
	lensType middleLensType = ml.getRefLensType(0, MIDDLE);
	lensType rightLensType = ml.getRefLensType(0, RIGHT);
	Mat leftMultiViewImage = multiViewImage[leftLensType][centerView.x][centerView.y];
	Mat middleMultiViewImage = multiViewImage[middleLensType][centerView.x][centerView.y];
	if (leftMultiViewImage.channels() != 1)
		cvtColor(leftMultiViewImage, leftMultiViewImage, CV_RGB2GRAY);
	if (middleMultiViewImage.channels() != 1)
		cvtColor(middleMultiViewImage, middleMultiViewImage, CV_RGB2GRAY);
	Mat shiftImage(middleMultiViewImage.size(), CV_8UC1);
	Mat diffImage;

	//Trimming LEFT Multi-view Image to Prevent Error from Edge Micro Lens
	Mat leftMultiViewRoiImage = Mat::zeros(leftMultiViewImage.size(), CV_8UC1);
	Point firstPos = ml.subGridRefPos[0][middleLensType], endPos;
	endPos.x = leftMultiViewImage.cols - firstPos.x;
	endPos.y = leftMultiViewImage.rows - firstPos.y;
	Rect roi(firstPos, endPos);
	leftMultiViewImage(roi).copyTo(leftMultiViewRoiImage(roi));

	//Calculate Lens Type Offset (where LEFT-MIDDLE Offset = RIGHT-MIDDLE Offset)
	//int diffNonZeroCnt;
	//int minDiffCnt = shiftImage.cols * shiftImage.rows;
	//for (int shift = 1; shift < (int)ml.diameter; shift++)
	//{
	//	getShiftImage(leftMultiViewRoiImage, shiftImage, Point(shift, 0));
	//	absdiff(middleMultiViewImage, shiftImage, diffImage);
	//	diffNonZeroCnt = countNonZero(diffImage);
	//	if (minDiffCnt > diffNonZeroCnt)
	//	{
	//		minDiffCnt = diffNonZeroCnt;
	//		lensTypeOffset = shift;
	//	}
	//}
	//Calculate Lens Type Offset (SAD)
	Scalar minSAD(DBL_MAX, DBL_MAX, DBL_MAX);
	for (int shift = 1; shift < (int)ml.diameter; shift++)
	{
		getShiftImage(leftMultiViewRoiImage, shiftImage, Point(shift, 0));
		absdiff(middleMultiViewImage, shiftImage, diffImage);
		Scalar curSAD = sum(diffImage);
		if (minSAD[0] > curSAD[0])
		{
			minSAD[0] = curSAD[0];
			lensTypeOffset = shift;
		}
	}
	//Calculate Lens Type Offset (NCC)
	//Scalar maxNCC(0, 0, 0);
	//Scalar mmVal = sum(middleMultiViewImage.mul(middleMultiViewImage));
	//for (int shift = 1; shift < (int)ml.diameter; shift++)
	//{
	//	getShiftImage(leftMultiViewRoiImage, shiftImage, Point(shift, 0));
	//	Scalar msVal = sum(middleMultiViewImage.mul(shiftImage));
	//	Scalar ssVal = sum(shiftImage.mul(shiftImage));
	//	Scalar curNCC = msVal / (mmVal * ssVal);
	//	if (maxNCC[0] < curNCC[0])
	//	{
	//		maxNCC[0] = curNCC[0];
	//		lensTypeOffset = shift;
	//	}
	//}
	//Calculate Lens Type Offset (ZNCC)
	//Scalar maxZNCC(0, 0, 0);
	//Scalar mMean = mean(middleMultiViewImage);
	//Scalar mmVal = sum(Mat(middleMultiViewImage - mMean).mul(middleMultiViewImage - mMean));
	//for (int shift = 1; shift < (int)ml.diameter; shift++)
	//{
	//	getShiftImage(leftMultiViewRoiImage, shiftImage, Point(shift, 0));
	//	Scalar sMean = mean(shiftImage);
	//	Scalar msVal = sum(Mat(middleMultiViewImage- mMean).mul(shiftImage - sMean));
	//	Scalar ssVal = sum(Mat(shiftImage - sMean).mul(shiftImage - sMean));
	//	Scalar curZNCC = msVal / (mmVal * ssVal);
	//	if (maxZNCC[0] < curZNCC[0])
	//	{
	//		maxZNCC[0] = curZNCC[0];
	//		lensTypeOffset = shift;
	//	}
	//}

	//Shift Multi-view Image
	Point view;
	for (int type = TYPE1; type < TYPE_ALL; type++)
	{
		if (leftLensType == type)
			getShiftImage(edgeStdMap[type], edgeStdMap[type], Point(lensTypeOffset, 0));
		else if (rightLensType == type)
			getShiftImage(edgeStdMap[type], edgeStdMap[type], Point(-lensTypeOffset, 0));

		for (view.x = 0; view.x < imageNum; view.x++)
		{
			for (view.y = 0; view.y < imageNum; view.y++)
			{
				if (leftLensType == type)
					getShiftImage(multiViewImage[type][view.x][view.y], multiViewImage[type][view.x][view.y], Point(lensTypeOffset, 0));
				else if (rightLensType == type)
					getShiftImage(multiViewImage[type][view.x][view.y], multiViewImage[type][view.x][view.y], Point(-lensTypeOffset, 0));
			}
		}
	}
}

void MultiViewImage::merge(vector<vector<Mat>>& destImage, int mmode, int lmode)
{
	//Allocation destImage Memory
	Point view;
	lensType middleLensType = ml.getRefLensType(0, MIDDLE);
	destImage.resize(imageNum);
	for (view.x = 0; view.x < imageNum; view.x++)
	{
		destImage[view.x].resize(imageNum);
		for (view.y = 0; view.y < imageNum; view.y++)
		{
			destImage[view.x][view.y] = Mat::zeros(multiViewImageSize, rawImage.type());
		}
	}

	//Merge
	if (mmode == AVERAGE)	//Simple AVERAGE Merge
	{
		for (view.x = 0; view.x < imageNum; view.x++)
			for (view.y = 0; view.y < imageNum; view.y++)
				for (int type = TYPE1; type < TYPE_ALL; type++)
					destImage[view.x][view.y] += (1.0 / 3.0) * multiViewImage[type][view.x][view.y];
	}
	else if (mmode == GAUSSIAN)	//GAUSSIAN Distribution Merge
	{
		//Calculate Alpha for Merge
		double patchMin = patchSizeList.front();
		double patchMax = patchSizeList.back();
		//patchMin = 1;
		//patchMax = (int)ml.diameter;
		double patchMid = (double)(patchMax - patchMin) * 0.5;
		vector<double> patchAlphaList[TYPE_ALL];
		for (int type = TYPE1; type < TYPE_ALL; type++)
		{
			patchAlphaList[type].resize((int)ml.diameter, 0);
			for (int patchCnt = 0; patchCnt < patchAlphaList[type].size(); patchCnt++)
			{
				double mu;
				if (lmode == NMF)
				{
					mu = patchMin + patchMid * type;
				}
				else if (lmode == FMN)
				{
					mu = patchMin + patchMid * (2 - type);
				}
				else if (lmode == NFM)
				{
					if (type == 0)
						mu = patchMin;
					else if (type == 1)
						mu = patchMin + patchMid * 2;
					else if (type == 2)
						mu = patchMin + patchMid;
				}
				else if (lmode == FNM)
				{
					if (type == 0)
						mu = patchMin + patchMid * 2;
					else if (type == 1)
						mu = patchMin;
					else if (type == 2)
						mu = patchMin + patchMid;
				}
				double sigma = patchMid * 0.5; //Within 96%
				double y = 1 / (sqrt(2 * PI) * sigma) * exp(-pow(patchCnt - mu, 2) / (2 * pow(sigma, 2)));
				patchAlphaList[type][patchCnt] = y;
			}
		}
		for (int patchCnt = 0; patchCnt < patchAlphaList[TYPE1].size(); patchCnt++)
		{
			double denominator = patchAlphaList[TYPE1][patchCnt] + patchAlphaList[TYPE2][patchCnt] + patchAlphaList[TYPE3][patchCnt];
			for (int type = TYPE1; type < TYPE_ALL; type++)
			{
				patchAlphaList[type][patchCnt] /= denominator;
			}
		}

		//Alpha Visualization
		if (false)
		{
			system("cls");
			for (int type = TYPE1; type < TYPE_ALL; type++)
			{
				cout << "=====Lens Type" << type + 1 << " Alpha Graph=====" << endl;
				for (int patchCnt = 0; patchCnt < patchAlphaList[type].size(); patchCnt++)
				{
					for (int i = 0; i < patchAlphaList[type][patchCnt] * 20; i++)
						cout << "*";
					cout << endl;
				}
				cout << endl;
			}
		}

		//Merge (where Take only middleLensType into Account)
		int patch;
		double alpha;
		for (view.x = 0; view.x < imageNum; view.x++)
		{
			for (view.y = 0; view.y < imageNum; view.y++)
			{
				for (int x = 0; x < multiViewImage[middleLensType][centerView.x][centerView.y].cols; x++)
				{
					for (int y = 0; y < multiViewImage[middleLensType][centerView.x][centerView.y].rows; y++)
					{
						double sum_alpha = 0;
						for (int type = TYPE1; type < TYPE_ALL; type++)
						{
							patch = patchMap[type].at<uchar>(y, x);
							sum_alpha += patchAlphaList[type][patch];
						}

						//patch = patchMap[middleLensType].at<uchar>(y, x);
						for (int type = TYPE1; type < TYPE_ALL; type++)
						{
							patch = patchMap[type].at<uchar>(y, x);
							alpha = patchAlphaList[type][patch] / sum_alpha;
							if (rawImage.channels() == 1)
								destImage[view.x][view.y].at<uchar>(y, x) += (uchar)(alpha * multiViewImage[type][view.x][view.y].at<uchar>(y, x));
							else
								destImage[view.x][view.y].at<Vec3b>(y, x) += (Vec3b)(alpha * multiViewImage[type][view.x][view.y].at<Vec3b>(y, x));
						}
					}
				}
			}
		}
	}
	else if (mmode == BLEND)
	{
		vector<vector<Mat>> tempDestImage(imageNum);
		for (view.x = 0; view.x < imageNum; view.x++)
		{
			tempDestImage[view.x].resize(imageNum);
			for (view.y = 0; view.y < imageNum; view.y++)
			{
				tempDestImage[view.x][view.y] = Mat::zeros(multiViewImageSize, rawImage.type());
			}
		}
		for (view.x = 0; view.x < imageNum; view.x++)
			for (view.y = 0; view.y < imageNum; view.y++)
				for (int type = TYPE1; type < TYPE_ALL; type++)
					tempDestImage[view.x][view.y] += (1.0 / 3.0) * multiViewImage[type][view.x][view.y];

		//Calculate Alpha for Merge
		double patchMin = patchSizeList.front();
		double patchMax = patchSizeList.back();
		//patchMin = 1;
		//patchMax = (int)ml.diameter;
		double patchMid = (double)(patchMax - patchMin) * 0.5;
		vector<double> patchAlphaList[TYPE_ALL];
		for (int type = TYPE1; type < TYPE_ALL; type++)
		{
			patchAlphaList[type].resize((int)ml.diameter, 0);
			for (int patchCnt = 0; patchCnt < patchAlphaList[type].size(); patchCnt++)
			{
				double mu;
				if (lmode == NMF)
				{
					mu = patchMin + patchMid * type;
				}
				else if (lmode == FMN)
				{
					mu = patchMin + patchMid * (2 - type);
				}
				else if (lmode == NFM)
				{
					if (type == 0)
						mu = patchMin;
					else if (type == 1)
						mu = patchMin + patchMid * 2;
					else if (type == 2)
						mu = patchMin + patchMid;
				}
				else if (lmode == FNM)
				{
					if (type == 0)
						mu = patchMin + patchMid * 2;
					else if (type == 1)
						mu = patchMin;
					else if (type == 2)
						mu = patchMin + patchMid;
				}
				double sigma = patchMid * 0.5; //Within 96%
				double y = 1 / (sqrt(2 * PI) * sigma) * exp(-pow(patchCnt - mu, 2) / (2 * pow(sigma, 2)));
				patchAlphaList[type][patchCnt] = y;
			}
		}
		for (int patchCnt = 0; patchCnt < patchAlphaList[TYPE1].size(); patchCnt++)
		{
			double denominator = patchAlphaList[TYPE1][patchCnt] + patchAlphaList[TYPE2][patchCnt] + patchAlphaList[TYPE3][patchCnt];
			for (int type = TYPE1; type < TYPE_ALL; type++)
			{
				patchAlphaList[type][patchCnt] /= denominator;
			}
		}

		//Alpha Visualization
		if (false)
		{
			system("cls");
			for (int type = TYPE1; type < TYPE_ALL; type++)
			{
				cout << "=====Lens Type" << type + 1 << " Alpha Graph=====" << endl;
				for (int patchCnt = 0; patchCnt < patchAlphaList[type].size(); patchCnt++)
				{
					for (int i = 0; i < patchAlphaList[type][patchCnt] * 20; i++)
						cout << "*";
					cout << endl;
				}
				cout << endl;
			}
		}

		//Merge (where Take only middleLensType into Account)
		int patch;
		double alpha;
		for (view.x = 0; view.x < imageNum; view.x++)
		{
			for (view.y = 0; view.y < imageNum; view.y++)
			{
				for (int x = 0; x < multiViewImage[middleLensType][centerView.x][centerView.y].cols; x++)
				{
					for (int y = 0; y < multiViewImage[middleLensType][centerView.x][centerView.y].rows; y++)
					{
						patch = patchMap[middleLensType].at<uchar>(y, x);
						for (int type = TYPE1; type < TYPE_ALL; type++)
						{
							alpha = patchAlphaList[type][patch];
							if (rawImage.channels() == 1)
								destImage[view.x][view.y].at<uchar>(y, x) += (uchar)(alpha * multiViewImage[type][view.x][view.y].at<uchar>(y, x));
							else
								destImage[view.x][view.y].at<Vec3b>(y, x) += (Vec3b)(alpha * multiViewImage[type][view.x][view.y].at<Vec3b>(y, x));
						}
					}
				}
			}
		}
		Mat map = Mat::zeros(edgeStdMap[0].size(), CV_32F);
		for (int type = TYPE1; type < TYPE_ALL; type++)
		{
			Mat temp;
			edgeStdMap[type].convertTo(temp, CV_32F, 1 / 255.f);
			map += temp / 3;
		}
		if (destImage[0][0].channels() == 3)
		{
			vector<Mat> bgr_map(3);
			map.copyTo(bgr_map[0]);
			map.copyTo(bgr_map[1]);
			map.copyTo(bgr_map[2]);
			cv::merge(bgr_map, map);
		}
		for (view.x = 0; view.x < imageNum; view.x++)
		{
			for (view.y = 0; view.y < imageNum; view.y++)
			{
				Mat gauss, avg;
				destImage[view.x][view.y].convertTo(gauss, CV_MAKETYPE(CV_32F, destImage[0][0].channels()));
				tempDestImage[view.x][view.y].convertTo(avg, CV_MAKETYPE(CV_32F, destImage[0][0].channels()));
				Mat tempDest = gauss.mul(2*map) + avg.mul(Scalar(1, 1, 1) - 2*map);
				tempDest.convertTo(destImage[view.x][view.y], CV_MAKETYPE(CV_8U, destImage[0][0].channels()));
			}
		}
	}
}

void MultiViewImage::correctionImage(vector<vector<Mat>>& destImage, bool isTrimming, bool isRotate)
{
	Point view;

	if (isTrimming)
	{
		lensType middleLensType = ml.getRefLensType(0, MIDDLE);
		Point2d leftUpPos;
		Size trimmingSize;
		leftUpPos.x = ml.subGridRefPos[0][middleLensType].x;
		leftUpPos.y = ml.subGridRefPos[0][middleLensType].y;
		trimmingSize.width = int((multiViewImageSize.width - (leftUpPos.x * 2)) / 8) * 8;	//For Octree
		trimmingSize.height = int((multiViewImageSize.height - (leftUpPos.y * 2)) / 8) * 8;	//For Octree

		for (view.x = 0; view.x < imageNum; view.x++)
		{
			for (view.y = 0; view.y < imageNum; view.y++)
			{
				destImage[view.x][view.y] = destImage[view.x][view.y](Rect2d(leftUpPos, trimmingSize));
			}
		}
	}

	if (isRotate && (abs(ml.rotation) > (PI / 4.0)))
	{
		for (view.x = 0; view.x < imageNum; view.x++)
		{
			for (view.y = 0; view.y < imageNum; view.y++)
			{
				Mat tempImage(Size(multiViewImageSize.height, multiViewImageSize.width), rawImage.type());
				transpose(destImage[view.x][view.y], tempImage);
				flip(tempImage, destImage[view.x][view.y], 0);
			}
		}
	}
}

void MultiViewImage::correctionImage(Mat& destImage, bool isTrimming, bool isRotate)
{
	Point view;

	if (isTrimming)
	{
		lensType middleLensType = ml.getRefLensType(0, MIDDLE);
		Point2d leftUpPos;
		Size trimmingSize;
		leftUpPos.x = ml.subGridRefPos[0][middleLensType].x;
		leftUpPos.y = ml.subGridRefPos[0][middleLensType].y;
		trimmingSize.width = int((multiViewImageSize.width - (leftUpPos.x * 2)) / 8) * 8;	//For Octree
		trimmingSize.height = int((multiViewImageSize.height - (leftUpPos.y * 2)) / 8) * 8;	//For Octree

		destImage = destImage(Rect2d(leftUpPos, trimmingSize));
	}

	if (isRotate && (abs(ml.rotation) >(PI / 4.0)))
	{
		for (view.x = 0; view.x < imageNum; view.x++)
		{
			for (view.y = 0; view.y < imageNum; view.y++)
			{
				Mat tempImage(Size(multiViewImageSize.height, multiViewImageSize.width), rawImage.type());
				transpose(destImage, tempImage);
				flip(tempImage, destImage, 0);
			}
		}
	}
}

void MultiViewImage::visualMode()
{
	while (waitKey(1) != 'v')
	{
		cv::imshow("rawImage", rawImage);
		//imshow("resizeRawImage", resizeRawImageList[0][getPatchListIndex(trackbarPatchSize)]);
		ml.draw(rawImage, TYPE_ALL);
		//ml.draw(resizeRawImageList[0][getPatchListIndex(trackbarPatchSize)], TYPE_ALL, trackbarPatchSize);
		patchVisualization((lensType)trackbarLensType);
		cv::imshow("multiViewImage[LensType][View.x][View.y]", multiViewImage[trackbarLensType][trackbarViewNum.x][trackbarViewNum.y]);
	}
}

void MultiViewImage::savePatchMap(char* folderName)
{
	char fullPath[128];
	char fullPath2[128];

	patchVisualization();
	for (int type = TYPE1; type < TYPE_ALL; type++)
	{
		correctionImage(patchVisualizationMap[type]);
		correctionImage(edgeStdVisualizationMap[type]);

		sprintf(fullPath, "%s/patchMap_%d.png", folderName, type + 1);
		sprintf(fullPath2, "%s/stdMap_%d.png", folderName, type + 1);
		imwrite(fullPath, patchVisualizationMap[type]);
		imwrite(fullPath2, edgeStdVisualizationMap[type]);
	}
}

void MultiViewImage::saveEachLensTypeCenterView(char* folderName)
{
	char fullPath[128];

	for (int type = TYPE1; type < TYPE_ALL; type++)
	{
		correctionImage(multiViewImage[type][centerView.x][centerView.y]);

		sprintf(fullPath, "%s/LensType_%d.png", folderName, type + 1);
		imwrite(fullPath, multiViewImage[type][centerView.x][centerView.y]);		
	}
}

void MultiViewImage::saveMultiViewImage_RGB(char* folderName, vector<vector<Mat>>& srcImage)
{
	char fullPath[128];

	Point view;
	for (view.y = 0; view.y < imageNum; view.y++)
	{
		for (view.x = 0; view.x < imageNum; view.x++)
		{
			sprintf(fullPath, "%s/image_%03d.png", folderName, view.y * imageNum + view.x + 1);
			//sprintf(fullPath, "%s/image_%02d%02d.png", newFolderName, view.x, view.y);

			imwrite(fullPath, srcImage[view.x][view.y]);
		}
	}
}

Mat MultiViewImage::colorCorrection(Mat src, bool isHistEqualization, float gamma)
{
	Mat dst(src.size(), CV_8UC3);
	if(isHistEqualization)
	{
		vector<Mat> channels(3);
		vector<Mat> channels_eq(3);
		split(src, channels);
			
		Ptr<CLAHE> clahe = createCLAHE();
		clahe->setClipLimit(2.);
		clahe->setTilesGridSize(Size(8, 8));

		clahe->apply(channels.at(0), channels_eq.at(0));
		clahe->apply(channels.at(1), channels_eq.at(1));
		clahe->apply(channels.at(2), channels_eq.at(2));
			
		//equalizeHist(channels.at(0), channels_eq.at(0));
		//equalizeHist(channels.at(1), channels_eq.at(1));
		//equalizeHist(channels.at(2), channels_eq.at(2));

		cv::merge(channels_eq, dst);
	}
	GammaCorrection(dst, dst, gamma);
	return dst;
}
void MultiViewImage::GammaCorrection(Mat& src, Mat& dst, float fGamma) 
{
	unsigned char lut[256];
	for (int i = 0; i < 256; i++)
	{
		lut[i] = saturate_cast<uchar>(pow((float)(i / 255.0), fGamma) * 255.0f);
	}
	dst = src.clone();
	const int channels = dst.channels();

	switch (channels)
	{
		case 1:
		{
			MatIterator_<uchar> it, end;
			for (it = dst.begin<uchar>(), end = dst.end<uchar>(); it != end; it++)
				*it = lut[(*it)];
			break;
		}
		case 3:
		{
			MatIterator_<Vec3b> it, end;
			for (it = dst.begin<Vec3b>(), end = dst.end<Vec3b>(); it != end; it++)
			{
				(*it)[0] = lut[((*it)[0])];
				(*it)[1] = lut[((*it)[1])];
				(*it)[2] = lut[((*it)[2])];
			}
			break;
		}
	} 
}

void MultiViewImage::saveMultiViewImage_YUV(char* folderName, vector<vector<Mat>>& srcImage)
{
	char fullPath[128];

	Point view;
	int w=srcImage[0][0].cols;
	int h=srcImage[0][0].rows;
	sprintf(fullPath, "%s/data.yuv", folderName);

	FILE* pFileOut = fopen(fullPath, "wb");
	int bufLen = w*h*3/2;
	unsigned char* pYuvBuf = new unsigned char[bufLen]; 

	for (view.y = 0; view.y < imageNum; view.y++)
	{
		for (view.x = 0; view.x < imageNum; view.x++)
		{
			//sprintf(fullPath, "%s/image_%03d.png", folderName, view.y * imageNum + view.x + 1);
			//sprintf(fullPath, "%s/image_%02d%02d.png", newFolderName, view.x, view.y);

			cvtColor(srcImage[view.x][view.y], srcImage[view.x][view.y], COLOR_BGR2YUV_I420);
			//cvtColor(srcImage[view.x][view.y], srcImage[view.x][view.y], COLOR_BGR2YUV_I420);
			//cvtColor(srcImage[view.x][view.y], srcImage[view.x][view.y], COLOR_YUV2BGRA_I420);
			memcpy(pYuvBuf, srcImage[view.x][view.y].data, bufLen*sizeof(unsigned char));
			fwrite(pYuvBuf, bufLen*sizeof(unsigned char), 1, pFileOut);

			//imwrite(fullPath, srcImage[view.x][view.y]);
		}
	}
	fclose(pFileOut);
}

void MultiViewImage::operator()(Mat srcImage, vector<vector<Mat>>& destImage, int pmode, int mmode, int lmode)
{
	//Set Input Image
	rawImage = srcImage.clone();
	Mat affineMatrix = getRotationMatrix2D(ml.center, -ml.rotation * 180 / PI, 1.0);
	warpAffine(rawImage, rawImage, affineMatrix, rawImage.size());
	generateResizeRawImage();

	for (int patchCnt = 0; patchCnt < patchSizeList.size(); patchCnt++)
	{
		rendering(true, true, patchSizeList[patchCnt]);
		for (int type = 0; type < 3; type++)
		{
			char str[256];
			sprintf(str, "./test_images/type%02d_ps%02d.png", type, patchCnt);
			imwrite(str, multiViewImage[type][0][0]);
		}
	}

	//Patch Estimation
	switch (pmode)
	{
	case LAPLACIAN_BASE_ALL:
		laplacianBaseEstimation(false);
		break;
	case LAPLACIAN_BASE_CENTER:
	    laplacianBaseEstimation(true, centerView);
		break;
	case DEPTH_BASE_V1:
		depthBaseEstimationV1();
		break;
	case DEPTH_BASE_V2:
		depthBaseEstimationV2();
		break;
	default:
		break;
	}

	//Generate Patch Map and Visualization
	generatePatchMap(pmode);

	//Generate Images by Rearranging Sub Image which is extracted from Raw Image
	//if (pmode != DEPTH_BASE_V3)
		rendering(false, false);

	//Shift Multi-view Image by Normalized Cross-correlation
	shiftMultiViewImage();

	//Merge Each Lens Type Multi-view Image (Set Output Image)
	merge(destImage, mmode, lmode);

	//Image Correction
	correctionImage(destImage);

	//Visual Mode
	if (waitKey(1) == 'v')
		visualMode();

	//Save Image for Debug
	//saveImage();
}