/* StereoMatching.cpp ======================================================================*

*Important
===================================================
Raytrix Lenslet Converting Version 0.2

Copyright: ONLY Available for ACADEMIC and MPEG Standardization Purposes
Production: Nagoya University


*Description
===================================================
Given a set of Raw video captured by multi-focused plenoptic camera (i.e. Ratrix camera, https://raytrix.de/) 
and its microlens calibration data, the provided source code can render multi-view video.

This tool is implemented by:

Fujii Laboratory
Department of Information and Communication Engineering
Graduate School of Engineering
Nagoya University


*Term of Use
===================================================
Term of Use: ANY kind of academic publication, and report using this tool MUST refer to the following TWO references.

[1] Shu Fujita, Sho Mikawa, Mehrdad Teratani(Panahpourtehrani), Keita Takahashi, and Toshiaki Fujii 
“Extracting Multi-View Images from Multi-Focused Plenoptic Camera,” 
International Workshop on Advanced Image Technology (IWAIT)/International Forum on Medical Imaging in Asia (IFMIA), 
in SPIE Digital Library, Singapore, January 2019

[2] ] Mehrdad Teratani(Panahpourtehrani), Shu Fujita, Wenzhe Ouyang, Keita Takahashi, Toshiaki Fujii, 
"A 3-D Imaging System Using Multi-Focused Plenoptic Camera and Tensor Display," 
2018 International Conference on 3D Immersion (IC3D2018) in IEEE Xplore Digital Library, Brussels, Belgium, December 2018 

===================================================
* Class for generating raytrix multi-view image
*
* @file		RenderingRxMV > MultiViewImage.h
* @author	S. Mikawa and S. Fujita
* @date		2018/02/dd
*======================================================================*/
#include "StereoMatching.h"

#define CV_CAST_8U(t) (uchar)(!((t) & ~255) ? (t) : (t) > 0 ? 255 : 0)

RaytrixStereoMatching::RaytrixStereoMatching(MicroLens ml, int minDisparity, int maxDisparity, Size blockSize)
{
	this->ml = ml;
	this->minDisparity = minDisparity;
	this->blockSize = blockSize;
	disparityNum = maxDisparity - minDisparity + 1;
	DSI.resize(disparityNum);

	errorTruncate = 40;	//Under 255/6
	preFilterCap = 49;
}

void RaytrixStereoMatching::absDiffTruncateSSE(Mat& src1, Mat& src2, uchar thresh, Mat& destImage)
{
	const int size = src1.size().area();
	uchar* s1 = src1.ptr<uchar>(0);
	uchar* s2 = src2.ptr<uchar>(0);
	uchar* d = destImage.ptr<uchar>(0);

	const __m128i trunc = _mm_set1_epi8(thresh);
	for (int i = 0; i < size - 16; i += 16)
	{
		//boost function
		__m128i a = _mm_load_si128((__m128i*)(s1));
		__m128i b = _mm_load_si128((__m128i*)(s2));
		__m128i c = _mm_load_si128((__m128i*)(d));

		_mm_stream_si128((__m128i*)(d), _mm_add_epi8(c, _mm_min_epu8(_mm_add_epi8(_mm_subs_epu8(a, b), _mm_subs_epu8(b, a)), trunc)));

		s1 += 16;
		s2 += 16;
		d += 16;
	}

	int end = size % 16;
	for (int i = 0; i < end; i++)
	{
		uchar a = abs(*s1 - *s2);
		*d = min(a, thresh);

		s1++;
		s2++;
		d++;
	}
}

void RaytrixStereoMatching::prefilterXSobel(Mat& srcImage, Mat& destImage, int ftzero)
{
	Mat tmpImage = Mat::zeros(srcImage.size(), CV_8U);
	int x, y;
	const int OFS = 256 * 4, TABSZ = OFS * 2 + 256;

	Size size = srcImage.size();
	const uchar val0 = ftzero;

	const int step = 2 * srcImage.cols;
	uchar* srow1 = srcImage.ptr<uchar>(0);
	uchar* dptr0 = tmpImage.ptr<uchar>(0);
	uchar* dptr1 = dptr0 + tmpImage.step;

	for (y = 0; y < size.height - 1; y += 2)
	{
		const uchar* srow0 = y > 0 ? srow1 - srcImage.step : size.height > 1 ? srow1 + srcImage.step : srow1;
		const uchar* srow2 = y < size.height - 1 ? srow1 + srcImage.step : size.height > 1 ? srow1 - srcImage.step : srow1;
		const uchar* srow3 = y < size.height - 2 ? srow1 + srcImage.step * 2 : srow1;

		dptr0[0] = dptr0[size.width - 1] = dptr1[0] = dptr1[size.width - 1] = val0;
		x = 1;

		__m128i z = _mm_setzero_si128(), ftz = _mm_set1_epi16((short)ftzero),
			ftz2 = _mm_set1_epi8(CV_CAST_8U(ftzero * 2));
		for (; x <= size.width - 9; x += 8)
		{
			//cout << y << ", " << x << endl;
			__m128i c0 = _mm_unpacklo_epi8(_mm_loadl_epi64((__m128i*)(srow0 + x - 1)), z);
			__m128i c1 = _mm_unpacklo_epi8(_mm_loadl_epi64((__m128i*)(srow1 + x - 1)), z);
			__m128i d0 = _mm_unpacklo_epi8(_mm_loadl_epi64((__m128i*)(srow0 + x + 1)), z);
			__m128i d1 = _mm_unpacklo_epi8(_mm_loadl_epi64((__m128i*)(srow1 + x + 1)), z);

			d0 = _mm_sub_epi16(d0, c0);
			d1 = _mm_sub_epi16(d1, c1);

			__m128i c2 = _mm_unpacklo_epi8(_mm_loadl_epi64((__m128i*)(srow2 + x - 1)), z);
			__m128i c3 = _mm_unpacklo_epi8(_mm_loadl_epi64((__m128i*)(srow2 + x - 1)), z);
			__m128i d2 = _mm_unpacklo_epi8(_mm_loadl_epi64((__m128i*)(srow2 + x + 1)), z);
			__m128i d3 = _mm_unpacklo_epi8(_mm_loadl_epi64((__m128i*)(srow2 + x + 1)), z);

			d2 = _mm_sub_epi16(d2, c2);
			d3 = _mm_sub_epi16(d3, c3);

			__m128i v0 = _mm_add_epi16(d0, _mm_add_epi16(d2, _mm_add_epi16(d1, d1)));
			__m128i v1 = _mm_add_epi16(d1, _mm_add_epi16(d3, _mm_add_epi16(d2, d2)));
			v0 = _mm_packus_epi16(_mm_add_epi16(v0, ftz), _mm_add_epi16(v1, ftz));
			v0 = _mm_min_epu8(v0, ftz2);

			_mm_storel_epi64((__m128i*)(dptr0 + x), v0);
			_mm_storel_epi64((__m128i*)(dptr1 + x), _mm_unpackhi_epi64(v0, v0));
		}
		srow1 += step;
		dptr0 += step;
		dptr1 += step;
	}
	for (; y < size.height; y++)
	{
		uchar* dptr = tmpImage.ptr<uchar>(y);
		for (x = 0; x < size.width; x++)
			dptr[x] = val0;
	}
	tmpImage.copyTo(destImage);
}

void RaytrixStereoMatching::shiftImage(InputArray src_, OutputArray dest_, const int hShift, const int vShift)
{
	//If hShift >= 0, then the input image shifts to right.
	//If vShift >= 0, then the input image shifts to down.

	if (dest_.empty())
		dest_.create(src_.size(), src_.type());

	Mat srcImage = src_.getMat();
	Mat destImage = dest_.getMat();
	Mat temp(srcImage.size(), srcImage.type());

	const int step = (int)srcImage.step;
	const int height = srcImage.rows;

	if (hShift >= 0)	// right shift case
	{
		uchar* s = srcImage.ptr<uchar>(0);
		uchar* d = temp.ptr<uchar>(0);

		for (int j = 0; j < height; j++, s += step, d += step)
		{
			const uchar v = s[0];
			memset(d, v, hShift);
			memcpy(d + hShift, s, step - hShift);
		}
	}
	else // left shift case
	{
		uchar* s = srcImage.ptr<uchar>(0);
		uchar* d = temp.ptr<uchar>(0);

		for (int j = 0; j < height; j++, s += step, d += step)
		{
			const uchar v = s[step - 1];
			memcpy(d, s - hShift, step + hShift);
			memset(d + step + hShift, v, -hShift);
		}
	}

	if (vShift >= 0)	// down shift case
	{
		uchar* s = temp.ptr<uchar>(0);
		uchar* d = destImage.ptr<uchar>(0);

		for (int j = 0; j < vShift; j++, d += step)
			memcpy(d, s, step);
		memcpy(d, s, step * (height - vShift));
	}
	else // up shift case
	{
		uchar* s = temp.ptr<uchar>(0);
		uchar* d = destImage.ptr<uchar>(0);

		memcpy(d, s - step * vShift, step * (height + vShift));
		s += step * (height - 1);
		d += step * (height + vShift);
		for (int j = vShift; j < 0; j++, d += step)
			memcpy(d, s, step);
	}
}

void RaytrixStereoMatching::getMatchingCostSAD(Mat& t, Mat& r, Mat& destImage, const int hd, const int vd)
{
	if (destImage.empty())
		destImage = Mat::zeros(t.size(), CV_8U);
	Mat bufferGray = Mat::zeros(t.size(), t.type());

	shiftImage(r, bufferGray, hd, vd);
	absDiffTruncateSSE(t, bufferGray, errorTruncate, destImage);
}

void RaytrixStereoMatching::getWTA(vector<Mat>& dsi, Mat& destImage)
{
	const int imsize = destImage.size().area();

	for (int i = 0; i < disparityNum; i++)
	{
		const short d = minDisparity + i;

		short* disp = destImage.ptr<short>(0);
		uchar* cost = minCostMap.ptr<uchar>(0);
		uchar* pDSI = dsi[i].ptr<uchar>(0);

		for (int j = imsize; j--; pDSI++, cost++, disp++)
		{
			if (*pDSI < *cost)
			{
				*cost = *pDSI;
				*disp = d;
			}
		}
	}
}

void RaytrixStereoMatching::generateLensMask(Mat& srcImage, Mat& maskImage, int erosionRadius)
{
	maskImage = Mat::zeros(srcImage.size(), CV_8U);

	int shift = 4;
	Point lens;
	Point2d lensCenter;

	for (int type = TYPE1; type < TYPE_ALL; type++)
	{
		for (lens.y = 0; lens.y < ml.num.y; lens.y++)
		{
			for (lens.x = 0; lens.x < ml.num.x; lens.x++)
			{
				lensCenter.x = ml.subGridRefPos[lens.y % 2][type].x + ml.diameter * 3 * lens.x;
				lensCenter.y = ml.subGridRefPos[0][type].y + ml.radius * sqrt(3) * lens.y;

				circle(maskImage, (1 << shift) * lensCenter, (1 << shift) * (int)(ml.radius - erosionRadius * 0.5), Scalar::all(255), -1, 8, shift);
			}
		}
	}

	compare(maskImage, 0, maskImage, CMP_EQ);
}

void RaytrixStereoMatching::initialization(Mat& srcImage)
{
	generateLensMask(srcImage, lensMaskImage, 2);
	minCostMap.create(srcImage.size(), CV_8U);
	minCostMap.setTo(UCHAR_MAX);

	for (int i = 0; i < DSI.size(); i++)
		DSI[i] = Mat::zeros(srcImage.size(), CV_8U);
}

void RaytrixStereoMatching::operator()(Mat& srcImage, Mat& disparityImage)
{
	//Initialization
	initialization(srcImage);

	//Set srcImage to targetImage
	if (srcImage.channels() == 3)
		cvtColor(srcImage, targetImage, CV_BGR2GRAY);
	else
		srcImage.copyTo(targetImage);

	//XSobel filter is Applied
	//prefilterXSobel(targetImage, targetImage, preFilterCap);
	//Laplacian(srcImage, targetImage, srcImage.depth(), 1, 3, 0, BORDER_DEFAULT);

	//Cost Computation for Each View
	for (int k = 0; k < 6; k++)
	{
		double hd, vd;
		if (k == LEFT) //ReferenceImage: Left Lens
		{
			shiftImage(targetImage, referenceImage, (int)ml.diameter, 0);
			hd = -1.0;
			vd = 0.0;
		}
		else if (k == RIGHT) //ReferenceImage: Right Lens
		{
			shiftImage(targetImage, referenceImage, (int)-ml.diameter, 0);
			hd = 1.0;
			vd = 0.0;
		}
		else if (k == LEFT_TOP)
		{
			shiftImage(targetImage, referenceImage, (int)(ml.diameter * 0.5), (int)(sqrt(3) * ml.diameter * 0.5));
			hd = -0.5;
			vd = -sqrt(3) * 0.5;
		}
		else if (k == LEFT_BOTTOM)
		{
			shiftImage(targetImage, referenceImage, (int)(ml.diameter * 0.5), (int)(-sqrt(3) * ml.diameter * 0.5));
			hd = -0.5;
			vd = sqrt(3) * 0.5;
		}
		else if (k == RIGHT_TOP)
		{
			shiftImage(targetImage, referenceImage, (int)(-ml.diameter * 0.5), (int)(sqrt(3) * ml.diameter * 0.5));
			hd = 0.5;
			vd = -sqrt(3) * 0.5;
		}
		else if (k == RIGHT_BOTTOM)
		{
			shiftImage(targetImage, referenceImage, (int)(-ml.diameter * 0.5), (int)(-sqrt(3) * ml.diameter * 0.5));
			hd = 0.5;
			vd = sqrt(3) * 0.5;
		}

		//Cost Computation for Each Disparity
#pragma omp parallel for
		for (int i = 0; i < disparityNum; i++)
		{
			const int d = minDisparity + i;
			getMatchingCostSAD(targetImage, referenceImage, DSI[i], (int)(hd * d), (int)(vd * d));
		}
	}

	//Cost Aggregation
#pragma omp parallel for
	for (int i = 0; i < disparityNum; i++)
	{
		boxFilter(DSI[i], DSI[i], srcImage.depth(), blockSize);
	}

	//Calculation Disparity by Winner Take All
	targetImage.convertTo(disparityImage, CV_16S);
	getWTA(DSI, disparityImage);
	disparityImage.convertTo(disparityImage, CV_8U);
	disparityImage.setTo(0, lensMaskImage);
}